#include <linux/fs.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/gfp.h>
#include <linux/mm.h>
#include <asm/uaccess.h>

#define DRIVER_MAJOR 240

static char *addr;

struct page *driver_nopage( struct vm_area_struct *vma,
	unsigned long address, int *type )
{
	struct page *page;

	printk("driver_nopage called\n");
	page = virt_to_page( addr );
	get_page( page );
	return page;
}

static struct vm_operations_struct driver_vm_ops = {
	.fault = driver_nopage,
};

static int driver_mmap( struct file *instanz, struct vm_area_struct *vma )
{
	printk("driver_mmap called\n");
	vma->vm_flags |= VM_RESERVED;
	vma->vm_ops = &driver_vm_ops;
	return 0;
}

static ssize_t driver_read( struct file *instanz, char *user, size_t count,
	loff_t *offset )
{
	int not_copied, to_copy;

	printk("driver_read called\n");
	to_copy = min( strlen(addr)+1, count );
	not_copied=copy_to_user(user,addr,to_copy);
	return to_copy-not_copied;
}

static struct file_operations fops = {
	.owner= THIS_MODULE,
	.read = driver_read,
	.mmap = driver_mmap,
};

static int __init mod_init(void)
{
	if(register_chrdev(DRIVER_MAJOR, "Hello", &fops) != 0)
			goto out;
	addr = (char *)get_zeroed_page( GFP_USER );
	printk("mmap: page at %p\n", addr );
	if( addr==NULL )
			goto unreg;
	return 0;
unreg:
	unregister_chrdev(DRIVER_MAJOR,"Hello");
out:
	printk("register_chrdev failed!\n");
	return -EIO;
}

static void __exit mod_exit(void)
{
	free_pages( (unsigned long)addr, 0 ); // 2^0 = free 1 page
	unregister_chrdev(DRIVER_MAJOR,"Hello");
}

module_init( mod_init );
module_exit( mod_exit );
MODULE_LICENSE("GPL");
/* vim: set ts=4 aw ic ai tw=80 sw=4: */
